"""
Copyright 2013 Steven Diamond

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .cbc_conif import CBC
from .cplex_conif import CPLEX
from .cvxopt_conif import CVXOPT
from .diffcp_conif import DIFFCP
from .ecos_conif import ECOS
from .glop_conif import GLOP
from .glpk_conif import GLPK
from .glpk_mi_conif import GLPK_MI
from .gurobi_conif import GUROBI
from .mosek_conif import MOSEK
from .nag_conif import NAG
from .pdlp_conif import PDLP
from .scip_conif import SCIP
from .scipy_conif import SCIPY
from .scs_conif import SCS
from .xpress_conif import XPRESS
