
# THIS FILE IS GENERATED FROM CVXPY SETUP.PY
short_version = '1.2.3'
version = '1.2.3'
full_version = '1.2.3'
git_revision = '928b207'
commit_count = '0'
release = True
if not release:
    version = full_version
